@@ -1,4 +1,12 @@
-#ifndef USS
+/***********************************************************************
+*
+* ibmfloat.c - IBM floating point operations for the TI-990/12 computer.
+*
+* Changes:
+*   12/16/05   DGP   Original. Borrowed and hacked from Hercules.
+*
+***********************************************************************/
+
 /* IBMFLOAT.C   (c) Copyright Peter Kuschnerus, 2000-2003            */
 /*              ESA/390 Hex Floatingpoint Instructions               */
 /*              Borrowed from Hercules for TI990                     */
@@ -8,37 +16,13 @@
 /* described in the manual ESA/390 Principles of Operation.          */
 /*-------------------------------------------------------------------*/
 
+#include <stdlib.h>
+#include <math.h>
+
 #include "simdef.h"
 #include "ibmfloat.h"
 
-extern uint16 pcreg;	/* The program PC */
-extern uint16 opcreg;	/* The old program PC */
 extern uint16 statreg;	/* The program status register */
-extern uint16 wpreg;	/* The program Workspace Pointer */
-extern uint16 lights;	/* The panel lights */
-extern int breakaddr;   /* The break point address */
-extern long pancount;	/* Instructions executed for panel */
-extern unsigned long instcount;
-
-extern int run;
-extern int idle;
-extern int runled;
-extern int idleled;
-extern int model;
-extern int breakflag;
-extern int enablepanel;
-extern int mapenabled;
-extern int traceenable;
-extern int tracestart;
-extern int traceend;
-extern int tracelim[2];
-extern int deferint;
-extern uint8 mapcrudata;
-extern uint16 errcrudata;
-extern uint32 memlen;
-extern uint32 maplatch;
-extern char view[MAXVIEW][81];
-extern FILE *tracefd;
 
 /*-------------------------------------------------------------------*/
 /* Definitions                                                       */
@@ -1930,4 +1914,3 @@ ibm_subs (uint32 *m0, uint32 *m1)
 }
 
 /* end of float.c */
-#endif /* USS */
