@@ -7,7 +7,10 @@
 *
 ***********************************************************************/
 
-extern void clearline (void);
-extern void clearscreen (void);
-extern void screenposition (int , int);
+extern void screenposition (int, int, int);
+extern void clearline (int);
+extern void cleartoendline (int);
+extern void clearscreen (int);
+extern void cleartoendscreen (int);
+extern void screenattribute (int, int);
 
