@@ -4,6 +4,8 @@
 *
 * Changes:
 *   06/14/04   DGP   Original.
+*   11/07/08   DGP   Added Instruction tracing and disassembler.
+*   11/26/14   DGP   Improve FP Condition Code checking.
 *
 ***********************************************************************/
 
@@ -17,34 +19,13 @@
 #include "ibmfloat.h"
 
 extern uint16 pcreg;	/* The program PC */
-extern uint16 opcreg;	/* The old program PC */
 extern uint16 statreg;	/* The program status register */
-extern uint16 wpreg;	/* The program Workspace Pointer */
-extern uint16 lights;	/* The panel lights */
-extern int breakaddr;   /* The break point address */
-extern long pancount;	/* Instructions executed for panel */
-extern unsigned long instcount;
-
-extern int run;
-extern int idle;
-extern int runled;
-extern int idleled;
+extern uint32 mpcreg;	/* Mapped PC */
 extern int model;
-extern int breakflag;
-extern int enablepanel;
-extern int mapenabled;
-extern int traceenable;
-extern int tracestart;
-extern int traceend;
-extern int tracelim[2];
-extern int deferint;
-extern int traceit;
-extern uint16 mapcrudata;
-extern uint16 errcrudata;
-extern uint32 memlen;
-extern uint32 maplatch;
-extern char view[MAXVIEW][81];
-extern FILE *tracefd;
+
+extern int trcidx;
+extern int trcopr;
+extern TraceInst trcbuf[TRCBUFMAX];
 
 /***********************************************************************
 * getfloatreg - Get a float register value.
@@ -100,21 +81,17 @@ static uint32
 getfloat (uint16 sr, uint16 st, int inc, int srcdst)
 {
    uint16 sa;
-   uint32 sval;
+   uint32 sval = -1;
 
    switch (st)
    {
    case 0:
 
-      if (traceit && inc)
-	 fprintf (tracefd, " R%d", sr);
       sval = getfloatreg (sr);
       break;
 
    case 1:
 
-      if (traceit && inc)
-	 fprintf (tracefd, " *R%d", sr);
       sa = GETREG (sr);
       sval = ((uint32)GETMEM (sa & 0xFFFE, srcdst) << 16) | 
 	      (uint32)GETMEM ((sa+2) & 0xFFFE, srcdst);
@@ -122,15 +99,12 @@ getfloat (uint16 sr, uint16 st, int inc, int srcdst)
 
    case 2:
 
-      sa = GETINST;
-      if (traceit && inc)
+      sa = GETINST(TRUE);
+      if (inc)
       {
-	 if (sr == 0)
-	    fprintf (tracefd, " @>%X", sa);
-	 else
-	    fprintf (tracefd, " @>%X(%d)", sa, sr);
+	 trcbuf[trcidx].ops[trcopr++] = sa;
+	 INCPC(2);
       }
-      if (inc) pcreg += 2;
       if (sr != 0) sa = (int16)sa + (int16)GETREG (sr);
       sval = ((uint32)GETMEM (sa & 0xFFFE, srcdst) << 16) | 
 	      (uint32)GETMEM ((sa+2) & 0xFFFE, srcdst);
@@ -138,8 +112,6 @@ getfloat (uint16 sr, uint16 st, int inc, int srcdst)
 
    case 3:
 
-      if (traceit && inc)
-	 fprintf (tracefd, " *R%d+", sr);
       sa = GETREG (sr);
       sval = ((uint32)GETMEM (sa & 0xFFFE, srcdst) << 16) | 
 	      (uint32)GETMEM ((sa+2) & 0xFFFE, srcdst);
@@ -158,21 +130,17 @@ static t_uint64
 getdouble (uint16 sr, uint16 st, int inc, int srcdst)
 {
    uint16 sa;
-   t_uint64 sval;
+   t_uint64 sval = -1;
 
    switch (st)
    {
    case 0:
 
-      if (traceit && inc)
-	 fprintf (tracefd, " R%d", sr);
       sval = getdoublereg(sr);
       break;
 
    case 1:
 
-      if (traceit && inc)
-	 fprintf (tracefd, " *R%d", sr);
       sa = GETREG (sr);
       sval = ((t_uint64)GETMEM (sa & 0xFFFE, srcdst) << 48) | 
 	     ((t_uint64)GETMEM ((sa+2) & 0xFFFE, srcdst) << 32) |
@@ -182,15 +150,12 @@ getdouble (uint16 sr, uint16 st, int inc, int srcdst)
 
    case 2:
 
-      sa = GETINST;
-      if (traceit && inc)
+      sa = GETINST(TRUE);
+      if (inc)
       {
-	 if (sr == 0)
-	    fprintf (tracefd, " @>%X", sa);
-	 else
-	    fprintf (tracefd, " @>%X(%d)", sa, sr);
+	 trcbuf[trcidx].ops[trcopr++] = sa;
+	 INCPC(2);
       }
-      if (inc) pcreg += 2;
       if (sr != 0) sa = (int16)sa + (int16)GETREG (sr);
       sval = ((t_uint64)GETMEM (sa & 0xFFFE, srcdst) << 48) | 
 	     ((t_uint64)GETMEM ((sa+2) & 0xFFFE, srcdst) << 32) |
@@ -200,8 +165,6 @@ getdouble (uint16 sr, uint16 st, int inc, int srcdst)
 
    case 3:
 
-      if (traceit && inc)
-	 fprintf (tracefd, " *R%d+", sr);
       sa = GETREG (sr);
       sval = ((t_uint64)GETMEM (sa & 0xFFFE, srcdst) << 48) | 
 	     ((t_uint64)GETMEM ((sa+2) & 0xFFFE, srcdst) << 32) |
@@ -226,16 +189,12 @@ putfloat (uint16 dr, uint16 dt, uint32 dval, int srcdst)
    {
    case 0:
 
-      if (traceit)
-	 fprintf (tracefd, " R%d", dr);
       PUTREG (dr, (uint16)((dval >> 16) & 0xFFFF));
       PUTREG (dr+1, (uint16)(dval & 0xFFFF));
       break;
 
    case 1:
 
-      if (traceit)
-	 fprintf (tracefd, " *R%d", dr);
       da = GETREG (dr);
       PUTMEM (da, (uint16)((dval >> 16) & 0xFFFF), srcdst);
       PUTMEM (da+2, (uint16)(dval & 0xFFFF), srcdst);
@@ -243,15 +202,9 @@ putfloat (uint16 dr, uint16 dt, uint32 dval, int srcdst)
 
    case 2:
 
-      da = GETINST;
-      if (traceit)
-      {
-	 if (dr == 0)
-	    fprintf (tracefd, " @>%X", da);
-	 else
-	    fprintf (tracefd, " @>%X(%d)", da, dr);
-      }
-      pcreg += 2;
+      da = GETINST(TRUE);
+      trcbuf[trcidx].ops[trcopr++] = da;
+      INCPC(2);
       if (dr != 0) da = (int16)da + (int16)GETREG (dr);
       PUTMEM (da, (uint16)((dval >> 16) & 0xFFFF), srcdst);
       PUTMEM (da+2, (uint16)(dval & 0xFFFF), srcdst);
@@ -259,8 +212,6 @@ putfloat (uint16 dr, uint16 dt, uint32 dval, int srcdst)
 
    case 3:
 
-      if (traceit)
-	 fprintf (tracefd, " *R%d+", dr);
       da = GETREG (dr);
       PUTREG (dr, GETREG (dr) + 4);
       PUTMEM (da, (uint16)((dval >> 16) & 0xFFFF), srcdst);
@@ -283,8 +234,6 @@ putdouble (uint16 dr, uint16 dt, t_uint64 dval, int srcdst)
    {
    case 0:
 
-      if (traceit)
-	 fprintf (tracefd, " R%d", dr);
       PUTREG (dr, (uint16)((dval >> 48) & 0xFFFF));
       PUTREG (dr+1, (uint16)((dval >> 32) & 0xFFFF));
       PUTREG (dr+2, (uint16)((dval >> 16) & 0xFFFF));
@@ -293,8 +242,6 @@ putdouble (uint16 dr, uint16 dt, t_uint64 dval, int srcdst)
 
    case 1:
 
-      if (traceit)
-	 fprintf (tracefd, " *R%d", dr);
       da = GETREG (dr);
       PUTMEM (da, (uint16)((dval >> 48) & 0xFFFF), srcdst);
       PUTMEM (da+2, (uint16)((dval >> 32) & 0xFFFF), srcdst);
@@ -304,15 +251,9 @@ putdouble (uint16 dr, uint16 dt, t_uint64 dval, int srcdst)
 
    case 2:
 
-      da = GETINST;
-      if (traceit)
-      {
-	 if (dr == 0)
-	    fprintf (tracefd, " @>%X", da);
-	 else
-	    fprintf (tracefd, " @>%X(%d)", da, dr);
-      }
-      pcreg += 2;
+      da = GETINST(TRUE);
+      trcbuf[trcidx].ops[trcopr++] = da;
+      INCPC(2);
       if (dr != 0) da = (int16)da + (int16)GETREG (dr);
       PUTMEM (da, (uint16)((dval >> 48) & 0xFFFF), srcdst);
       PUTMEM (da+2, (uint16)((dval >> 32) & 0xFFFF), srcdst);
@@ -322,8 +263,6 @@ putdouble (uint16 dr, uint16 dt, t_uint64 dval, int srcdst)
 
    case 3:
 
-      if (traceit)
-	 fprintf (tracefd, " *R%d+", dr);
       da = GETREG (dr);
       PUTREG (dr, GETREG (dr) + 8);
       PUTMEM (da, (uint16)((dval >> 48) & 0xFFFF), srcdst);
@@ -372,6 +311,25 @@ cmpdoublezero (t_uint64 val)
    else CLR_EQ;
 }
 
+/***********************************************************************
+* checkfloatcc - Compare double value to zero.
+***********************************************************************/
+
+static void
+checkfloatcc (int cc)
+{
+   if (cc == PGM_EXPONENT_UNDERFLOW_EXCEPTION) 
+   {
+      CLR_CARRY;
+      SET_OVER;
+   }
+   else if (cc == PGM_EXPONENT_OVERFLOW_EXCEPTION)
+   {
+      SET_CARRY;
+      SET_OVER;
+   }
+}
+
 /***********************************************************************
 * proc6fltop - Process type 6 floating point operations.
 ***********************************************************************/
@@ -397,7 +355,6 @@ proc6fltop (uint16 inst)
    {
    case 1: /* AR */
 
-      if (traceit) fprintf (tracefd, "AR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = AR sr = %d, st = %d\n",
@@ -410,21 +367,19 @@ proc6fltop (uint16 inst)
 	       "   acc = >%08X\n", acc);
       fprintf (stderr,
 	       "   fop = >%08X\n", fop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_adds (&acc, &fop);
       putfloatreg (0, acc);
       cmpfloatzero (acc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   acc = >%08X\n", cc, acc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 2: /* CIR */
 
-      if (traceit) fprintf (tracefd, "CIR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CIR sr = %d, st = %d\n",
@@ -438,7 +393,6 @@ proc6fltop (uint16 inst)
 
    case 3: /* SR */
 
-      if (traceit) fprintf (tracefd, "SR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = SR sr = %d, st = %d\n",
@@ -451,21 +405,19 @@ proc6fltop (uint16 inst)
 	       "   acc = >%08X\n", acc);
       fprintf (stderr,
 	       "   fop = >%08X\n", fop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_subs (&acc, &fop);
       putfloatreg (0, acc);
       cmpfloatzero (acc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   acc = >%08X\n", cc, acc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 4: /* MR */
 
-      if (traceit) fprintf (tracefd, "MR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = MR sr = %d, st = %d\n",
@@ -478,21 +430,19 @@ proc6fltop (uint16 inst)
 	       "   acc = >%08X\n", acc);
       fprintf (stderr,
 	       "   fop = >%08X\n", fop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_mpys (&acc, &fop);
       putfloatreg (0, acc);
       cmpfloatzero (acc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   acc = >%08X\n", cc, acc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 5: /* DR */
 
-      if (traceit) fprintf (tracefd, "DR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = DR sr = %d, st = %d\n",
@@ -505,21 +455,19 @@ proc6fltop (uint16 inst)
 	       "   acc = >%08X\n", acc);
       fprintf (stderr,
 	       "   fop = >%08X\n", fop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_divs (&acc, &fop);
       putfloatreg (0, acc);
       cmpfloatzero (acc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   acc = >%08X\n", cc, acc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 6: /* LR */
 
-      if (traceit) fprintf (tracefd, "LR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = LR sr = %d, st = %d\n",
@@ -532,7 +480,6 @@ proc6fltop (uint16 inst)
 
    case 7: /* STR */
 
-      if (traceit) fprintf (tracefd, "STR");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = STR sr = %d, st = %d\n",
@@ -545,7 +492,6 @@ proc6fltop (uint16 inst)
 
    case 9: /* AD */
 
-      if (traceit) fprintf (tracefd, "AD");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = AD sr = %d, st = %d\n",
@@ -558,21 +504,19 @@ proc6fltop (uint16 inst)
 	       "   dacc = >%016llX\n", dacc);
       fprintf (stderr,
 	       "   dop = >%016llX\n", dop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_addl (&dacc, &dop);
       putdoublereg (0, dacc);
       cmpdoublezero (dacc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   dacc = >%016llX\n", cc, dacc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 10: /* CID */
 
-      if (traceit) fprintf (tracefd, "CID");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CID sr = %d, st = %d\n",
@@ -586,7 +530,6 @@ proc6fltop (uint16 inst)
 
    case 11: /* SD */
 
-      if (traceit) fprintf (tracefd, "SD");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = SD sr = %d, st = %d\n",
@@ -599,21 +542,19 @@ proc6fltop (uint16 inst)
 	       "   dacc = >%016llX\n", dacc);
       fprintf (stderr,
 	       "   dop = >%016llX\n", dop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_subl (&dacc, &dop);
       putdoublereg (0, dacc);
       cmpdoublezero (dacc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   dacc = >%016llX\n", cc, dacc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 12: /* MD */
 
-      if (traceit) fprintf (tracefd, "MD");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = MD sr = %d, st = %d\n",
@@ -626,21 +567,19 @@ proc6fltop (uint16 inst)
 	       "   dacc = >%016llX\n", dacc);
       fprintf (stderr,
 	       "   dop = >%016llX\n", dop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_mpyl (&dacc, &dop);
       putdoublereg (0, dacc);
       cmpdoublezero (dacc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   dacc = >%016llX\n", cc, dacc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 13: /* DD */
 
-      if (traceit) fprintf (tracefd, "DD");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = DD sr = %d, st = %d\n",
@@ -653,21 +592,19 @@ proc6fltop (uint16 inst)
 	       "   dacc = >%016llX\n", dacc);
       fprintf (stderr,
 	       "   dop = >%016llX\n", dop);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       cc = ibm_divl (&dacc, &dop);
       putdoublereg (0, dacc);
       cmpdoublezero (dacc);
+      checkfloatcc (cc);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "   cc = %d\n   dacc = >%016llX\n", cc, dacc);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       break;
 
    case 14: /* LD */
 
-      if (traceit) fprintf (tracefd, "LD");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = LD sr = %d, st = %d\n",
@@ -680,7 +617,6 @@ proc6fltop (uint16 inst)
 
    case 15: /* STD */
 
-      if (traceit) fprintf (tracefd, "STD");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = STD sr = %d, st = %d\n",
@@ -714,12 +650,10 @@ proc7fltop (uint16 inst)
    {
    case 0: /* CRI */
 
-      if (traceit) fprintf (tracefd, "CRI");
       val = getfloatreg (0);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CRI\n   bv = >%08X\n", val);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       CLR_CARRY;
       CLR_OVER;
@@ -736,19 +670,14 @@ proc7fltop (uint16 inst)
       cval &= 0xFFFF;
       PUTREG(0, cval);
       cmpwordzero (cval);
-#ifdef DEBUGFINSTR
-      fprintf (stderr,"   stat = >%04X\n", statreg);
-#endif
       break;
 
    case 1: /* CDI */
 
-      if (traceit) fprintf (tracefd, "CDI");
       dval = getdoublereg (0);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CDI\n   bv = >%016llX\n", dval);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       CLR_CARRY;
       CLR_OVER;
@@ -765,14 +694,10 @@ proc7fltop (uint16 inst)
       cval &= 0xFFFF;
       PUTREG(0, cval);
       cmpwordzero (cval);
-#ifdef DEBUGFINSTR
-      fprintf (stderr,"   stat = >%04X\n", statreg);
-#endif
       break;
 
    case 2: /* NEGR */
 
-      if (traceit) fprintf (tracefd, "NEGR");
       val = getfloatreg (0);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
@@ -790,7 +715,6 @@ proc7fltop (uint16 inst)
 
    case 3: /* NEGD */
 
-      if (traceit) fprintf (tracefd, "NEGD");
       dval = getdoublereg(0);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
@@ -808,12 +732,10 @@ proc7fltop (uint16 inst)
 
    case 4: /* CRE */
 
-      if (traceit) fprintf (tracefd, "CRE");
       val = getfloatreg (0);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CRE\n   bv = >%016llX\n", dval);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       CLR_CARRY;
       CLR_OVER;
@@ -824,19 +746,14 @@ proc7fltop (uint16 inst)
 #endif
       putfloatreg (0, cval);
       cmplongzero (cval);
-#ifdef DEBUGEINSTR
-      fprintf (stderr,"   stat = >%04X\n", statreg);
-#endif
       break;
 
    case 5: /* CDE */
 
-      if (traceit) fprintf (tracefd, "CDE");
       dval = getdoublereg (0);
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CDE\n   bv = >%016llX\n", dval);
-      fprintf (stderr,"   stat = >%04X\n", statreg);
 #endif
       CLR_CARRY;
       CLR_OVER;
@@ -847,14 +764,10 @@ proc7fltop (uint16 inst)
 #endif
       putfloatreg (0, cval);
       cmplongzero (cval);
-#ifdef DEBUGEINSTR
-      fprintf (stderr,"   stat = >%04X\n", statreg);
-#endif
       break;
 
    case 6: /* CER */
 
-      if (traceit) fprintf (tracefd, "CER");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CER\n");
@@ -867,7 +780,6 @@ proc7fltop (uint16 inst)
 
    case 7: /* CED */
 
-      if (traceit) fprintf (tracefd, "CED");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = CED\n");
@@ -881,7 +793,6 @@ proc7fltop (uint16 inst)
    case 14: /* XIT */
    case 15: /* XIT */
 
-      if (traceit) fprintf (tracefd, "XIT");
 #ifdef DEBUGFINSTR
       fprintf (stderr,
 	       "INST = XIT\n");
