@@ -10,6 +10,14 @@
 *   07/29/04   DGP   Added tapereset function.
 *   12/02/04   DGP   Correct tape read.
 *   12/05/04   DGP   Correct backspace.
+*   04/28/05   DGP   Changed to use switches.
+*   10/29/13   DGP   Correct status on unconnected devices.
+*   02/12/14   DGP   Fixed compressed (binary) load.
+*   03/12/15   DGP   Added real device support for tape.
+*   05/02/16   DGP   Don't set density, run with current setting.
+*   05/04/16   DGP   Added tapecheck() function.
+*   07/05/16   DGP   Fixed unit offline with interrupts enabled.
+*   02/21/19   DGP   MacOS Darwin changes
 *
 * Tape record format:
 *   4 bytes (little-endian): block length
@@ -22,12 +30,13 @@
 ***********************************************************************/
 
 #include <stdio.h>
+#include <unistd.h>
 #include <string.h>
 #include <memory.h>
 #include <ctype.h>
 #include <errno.h>
 
-#if defined (_WIN32)
+#if defined (WIN32)
 #define __TTYROUTINES 0
 #include <conio.h>
 #include <windows.h>
@@ -39,6 +48,10 @@
 #include <sys/ioctl.h>
 #include <termios.h>
 #include <pthread.h>
+#if !defined(DARWIN)
+#define HAVE_SYS_MTIO_H 1
+#include <sys/mtio.h>
+#endif
 #endif
 
 #include "simdef.h"
@@ -46,36 +59,20 @@
 extern uint16 pcreg;	/* The program PC */
 extern uint16 statreg;	/* The program status register */
 extern uint16 wpreg;	/* The program Workspace Pointer */
-extern uint16 lights;	/* The panel lights */
 
 extern int run;
-extern int idle;
-extern int fault;
 extern int devcnt;
-extern int model;
-extern int mapenabled;
-extern int pancount;
+extern int bootfromrom;
 extern uint32 memlen;
-extern char view[MAXVIEW][81];
+extern char view[MAXVIEW][MAXVIEWLEN+1];
 
-extern uint8 memory[MEMSIZE];
+extern uint8 memory[SYSMEMSIZE];
 extern Device devices[MAXDEVICES];
 
 /*
 ** Command words access macros
 */
 
-#define GETW0(d)    GETMEM0((d)->devaddr+TPCSSTART)
-#define GETW7(d)    GETMEM0((d)->devaddr+TPCSSTART+14)
-
-#define PUTW0(d,v)  PUTMEM0((d)->devaddr+TPCSSTART,v)
-#define PUTW7(d,v)  PUTMEM0((d)->devaddr+TPCSSTART+14,v)
-
-#define GETLEN(d)   GETMEM0((d)->devaddr+TPCSSTART+8)
-#define PUTLEN(d,v) PUTMEM0((d)->devaddr+TPCSSTART+8,v)
-
-#define GETADDR(d) \
-(GETMEMB0((d)->devaddr+TPCSSTART+13)<<16)|GETMEM0((d)->devaddr+TPCSSTART+10)
 #define GETCMD(d)   GETMEMB0((d)->devaddr+TPCSSTART+12) & 0xF
 #define GETUNIT(d)  (GETMEMB0((d)->devaddr+TPCSSTART+12) >> 4) & 0xF
 
@@ -109,6 +106,15 @@ extern Device devices[MAXDEVICES];
 #define FMEBIT	0x0002		/* Format error */
 #define TPEBIT	0x0001		/* Tape error */
 
+#if defined(UNIX) && defined(HAVE_SYS_MTIO_H)
+static struct mtop mt_weof   = { MTWEOF, 1 };
+static struct mtop mt_rew    = { MTREW, 1 };
+static struct mtop mt_bsr    = { MTBSR, 1 };
+static struct mtop mt_fsr    = { MTFSR, 1 };
+static struct mtop mt_setblk = { MTSETBLK, 0 };  /* blockize = 0 (variable) */
+/*static struct mtop mt_setden = { MTSETDENSITY, 0x02 };*/ /* density = 1600 */
+#endif
+
 /***********************************************************************
 * tapereadint - Read an integer.
 ***********************************************************************/
@@ -122,8 +128,8 @@ tapereadint (FILE *fd)
    r = r | (fgetc (fd) << 8);
    r = r | (fgetc (fd) << 16);
    r = r | (fgetc (fd) << 24);
-   if (feof(fd)) return (-2);
-   return (r & 0x7FFF);
+   if (feof (fd)) return (TAPEEOT);
+   return (r);
 }
 
 /***********************************************************************
@@ -140,43 +146,6 @@ tapewriteint (FILE *fd, int v)
    fputc ((v >> 24) & 0xFF, fd);
 }
 
-/***********************************************************************
-* tapeopen - Open tape.
-***********************************************************************/
-
-int 
-tapeopen (Device *dev)
-{
-#ifdef DEBUGTAPE
-   fprintf (stderr, "tapeopen: ENTERED\n");
-   fprintf (stderr, "   file = %s\n", dev->file);
-#endif
-
-   if (dev->fd == DEV_NULL)
-   {
-      PUTW0 (dev, OLBIT | WPBIT);
-      PUTW7 (dev, IDLEBIT); /* Mark idle */
-      return (0);
-   }
-
-   /*
-   ** Seek to beginning of the tape file.
-   */
-
-   if (fseek (dev->fd, 0, SEEK_SET) < 0)
-   {
-      sprintf (view[0], "tapeopen: seek failed: %s",
-	    strerror (ERRNO));
-      sprintf (view[1], "filename: %s", dev->file);
-      return (-1);
-   }
-
-   PUTW0 (dev, BOTBIT | (dev->info.tapeinfo.writeprotect ? WPBIT : 0));
-   PUTW7 (dev, IDLEBIT); /* Mark idle */
-
-   return (0);
-}
-
 /***********************************************************************
 * tapeputmem - Put data into memory.
 ***********************************************************************/
@@ -202,6 +171,442 @@ tapegetmem (uint32 ma)
    return (GETMEM0 (ma));
 }
 
+/***********************************************************************
+* tapesetchar - Set tape characteristics.
+***********************************************************************/
+
+static void 
+tapesetchar (Device *udev)
+{
+#if defined(UNIX) && defined(HAVE_SYS_MTIO_H)
+   if (udev->realdevice)
+   {
+      ioctl (udev->sfd, MTIOCTOP, &mt_setblk);
+      /*ioctl (udev->sfd, MTIOCTOP, &mt_setden);*/
+   }
+#endif
+}
+
+/***********************************************************************
+* taperewind - Rewind tape.
+***********************************************************************/
+
+static int 
+taperewind (Device *udev)
+{
+   int rc = 0;
+
+#ifdef DEBUGTAPE
+   fprintf (stderr, "taperewind: \n");
+#endif
+
+   udev->info.tapeinfo.curpos = 0;
+#if defined(UNIX) && defined(HAVE_SYS_MTIO_H)
+   if (udev->realdevice)
+      rc = ioctl (udev->sfd, MTIOCTOP, &mt_rew);
+   else
+#endif
+      rc = fseek (udev->infd, 0, SEEK_SET);
+
+   if (rc < 0)
+   {
+      sprintf (view[0], "Device %s: REWIND failed: %s",
+	       udev->name, strerror (ERRNO));
+      snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
+   }
+
+   return (rc);
+}
+
+/***********************************************************************
+* tapeweof - Write EOF on tape.
+***********************************************************************/
+
+static int 
+tapeweof (Device *udev)
+{
+   int rc = 0;
+
+#ifdef DEBUGTAPE
+   fprintf (stderr, "tapeweof: \n");
+#endif
+
+#if defined(UNIX) && defined(HAVE_SYS_MTIO_H)
+   if (udev->realdevice)
+      rc = ioctl (udev->sfd, MTIOCTOP, &mt_weof);
+   else
+#endif
+      tapewriteint (udev->infd, 0);
+
+   if (rc < 0)
+   {
+      sprintf (view[0], "Device %s: WRITEEOF failed: %s",
+	       udev->name, strerror (ERRNO));
+      snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
+   }
+
+   return (rc);
+}
+
+/***********************************************************************
+* tapebsr - Backspace record on tape.
+***********************************************************************/
+
+static int 
+tapebsr (Device *udev, int count)
+{
+   int rc = 0;
+
+#ifdef DEBUGTAPEBSR
+   fprintf (stderr, "tapebsr: count = %d, curpos = %ld\n",
+	    count, udev->info.tapeinfo.curpos);
+#endif
+
+   if (udev->info.tapeinfo.curpos <= 0)
+   {
+      udev->info.tapeinfo.curpos = 0;
+      count = 0;
+   }
+   else while (count && !rc)
+   {
+#if defined(UNIX) && defined(HAVE_SYS_MTIO_H)
+      if (udev->realdevice)
+      {
+	 rc = ioctl (udev->sfd, MTIOCTOP, &mt_bsr);
+      }
+      else
+#endif
+      {
+	 int reclen;
+
+#ifdef DEBUGTAPEBSR
+	 fprintf (stderr, "   cp0 = %ld\n", udev->info.tapeinfo.curpos);
+#endif
+	 udev->info.tapeinfo.curpos -= 4;
+#ifdef DEBUGTAPEBSR
+	 fprintf (stderr, "   cp1 = %ld\n", udev->info.tapeinfo.curpos);
+#endif
+
+	 if (fseek (udev->infd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
+	 {
+	    rc = TAPEERR;
+#ifdef DEBUGTAPEBSR
+	    perror ("fseek1 failed");
+#endif
+	    break;
+	 }
+	 reclen = tapereadint (udev->infd);
+#ifdef DEBUGTAPEBSR
+	 fprintf (stderr, "   reclen = %d\n", reclen);
+#endif
+	 if (reclen > 0)
+	    udev->info.tapeinfo.curpos -= reclen + 4;
+#ifdef DEBUGTAPEBSR
+	 fprintf (stderr, "   cp2 = %ld\n", udev->info.tapeinfo.curpos);
+#endif
+	 if (fseek (udev->infd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
+	 {
+#ifdef DEBUGTAPEBSR
+	    perror ("fseek2 failed");
+#endif
+	    rc = TAPEERR;
+	    break;
+	 }
+      }
+      count --;
+   }
+
+   if (rc < 0)
+   {
+      sprintf (view[0], "Device %s: BCKSKIP failed: %s",
+	       udev->name, strerror (ERRNO));
+      snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
+   }
+
+   PUTLEN (udev, count); 
+   return (rc);
+}
+
+/***********************************************************************
+* tapefsr - Forwardspace record on tape.
+***********************************************************************/
+
+static int 
+tapefsr (Device *udev, int count)
+{
+   int rc = 0;
+
+#ifdef DEBUGTAPE
+   fprintf (stderr, "tapefsr: count = %d\n", count);
+#endif
+
+   while (count && !rc)
+   {
+#if defined(UNIX) && defined(HAVE_SYS_MTIO_H)
+      if (udev->realdevice)
+      {
+	 rc = ioctl (udev->sfd, MTIOCTOP, &mt_fsr);
+	 if (rc < 0 && ERRNO == ENOSPC)
+	    rc = TAPEEOT;
+      }
+      else
+#endif
+      {
+	 int i;
+	 int reclen;
+
+	 reclen = tapereadint (udev->infd);
+	 if (reclen > 0)
+	 {
+	    for (i = 0; i < reclen; i++)
+	    {
+	       fgetc (udev->infd);
+	    }
+	    i = tapereadint (udev->infd);
+	    if (i != reclen)
+	    {
+	       rc = TAPEERR;
+	    }
+	 }
+	 else if (reclen == 0)
+	 {
+	    rc = TAPEEOF;
+	 }
+	 else if (reclen == -2)
+	 {
+	    rc = TAPEEOT;
+	 }
+      }
+      count--;
+   }
+
+   if (rc < 0)
+   {
+      sprintf (view[0], "Device %s: FWDSKIP failed: %s",
+	       udev->name, strerror (ERRNO));
+      snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
+   }
+
+   PUTLEN (udev, count); 
+   return (rc);
+}
+
+/***********************************************************************
+* taperead - Read a record from tape.
+***********************************************************************/
+
+static int 
+taperead (Device *udev, uint32 addr, int buflen)
+{
+   int rc = 0;
+   int reclen;
+   int i, j;
+
+
+#ifdef UNIX
+   if (udev->realdevice)
+   {
+     reclen = read (udev->sfd, udev->info.tapeinfo.record, TAPEMAXREC);
+   }
+   else
+#endif
+   {
+      reclen = tapereadint (udev->infd);
+      if (reclen > 0)
+      {
+	 fread (udev->info.tapeinfo.record, 1, reclen, udev->infd);
+	 i = tapereadint (udev->infd);
+	 if (i != reclen)
+	 {
+#ifdef DEBUGTAPE
+	    fprintf (stderr, "taperead: error: reclen = %d\n", reclen);
+	    fprintf (stderr, "   traillen = %d\n", i);
+#endif
+	    rc = TAPEERR;
+	 }
+      }
+   }
+
+#ifdef DEBUGTAPE
+   fprintf (stderr, "taperead: buflen = %d, reclen = %d\n", buflen, reclen);
+   if (!udev->realdevice)
+      fprintf (stderr, "   cp = %ld\n", udev->info.tapeinfo.curpos);
+#endif
+
+   if (!rc && reclen > 0)
+   {
+      for (j = i = 0; i < reclen;)
+      {
+	 uint16 d, c;
+
+	 d = udev->info.tapeinfo.record[i++];
+	 if (i < reclen)
+	 {
+	    c = udev->info.tapeinfo.record[i++];
+	 }
+	 else
+	    c = 0xFF;
+	 d = (d << 8) | c;
+	 if (j < buflen)
+	 {
+	    tapeputmem (addr, d);
+	    j += 2;
+	    addr += 2;
+	 }
+      }
+      if (i < buflen)
+      {
+	 buflen -= i;
+      }
+      else
+      {
+	 buflen = 0;
+      }
+      rc = buflen;
+
+#ifdef DEBUGTAPEDATA
+      HEXDUMP (stderr, &memory[(((GETADDR (udev)) + 1) & 0x1FFFFE)],
+	       j, (((GETADDR (udev)) + 1) & 0x1FFFFE));
+#endif
+      PUTLEN (udev, buflen); 
+      PUTADDR (udev, addr);
+   }
+   else if (reclen == 0)
+   {
+      rc = TAPEEOF;
+   }
+   else if (reclen == -2)
+   {
+      rc = TAPEEOT;
+   }
+
+   return (rc);
+}
+
+/***********************************************************************
+* tapewrite - Write a record to tape.
+***********************************************************************/
+
+static int 
+tapewrite (Device *udev, uint32 addr, int buflen)
+{
+   int rc = 0;
+   int i = 0;
+
+#ifdef DEBUGTAPE
+   fprintf (stderr, "tapewrite: buflen = %d\n", buflen);
+#endif
+   if (!(udev->switches & SWPROTECT))
+   {
+#ifdef DEBUGTAPEDATA
+      HEXDUMP (stderr, &memory[(((GETADDR (udev)) + 1) & 0x1FFFFE)],
+	       GETLEN(udev), (((GETADDR (udev)) + 1) & 0x1FFFFE));
+#endif
+      for (i = 0; i < buflen; )
+      {
+	 uint16 d;
+
+	 d = tapegetmem (addr);
+	 udev->info.tapeinfo.record[i++] =  (d >> 8) & 0xFF;
+	 if (i < buflen)
+	 {
+	    udev->info.tapeinfo.record[i++] = d & 0xFF;
+	 }
+	 addr += 2;
+      }
+
+#ifdef UNIX
+      if (udev->realdevice)
+      {
+         i = write (udev->sfd, udev->info.tapeinfo.record, buflen);
+      }
+      else
+#endif
+      {
+	 tapewriteint (udev->infd, buflen);
+	 i = fwrite (udev->info.tapeinfo.record, 1, buflen, udev->infd);
+	 tapewriteint (udev->infd, buflen);
+      }
+      if (i != buflen)
+	 rc = TAPEERR;
+   }
+   PUTLEN (udev, buflen-i);
+   PUTADDR (udev, addr);
+   return (rc);
+}
+
+/***********************************************************************
+* tapecheck - Check tape for valid format.
+***********************************************************************/
+
+static int 
+tapecheck (Device *dev)
+{
+   int rc;
+   int rlen;
+
+   if ((rc = taperewind (dev)) == 0)
+   {
+      rlen = tapereadint (dev->infd);
+      if ((rc = fseek (dev->infd, rlen + 4, SEEK_SET)) == 0)
+      {
+	 if (rlen != tapereadint(dev->infd))
+	    rc = -1;
+      }
+   }
+   if (rc < 0)
+   {
+      sprintf (view[0], "Device %s: Invalid tape format", dev->name);
+      snprintf (view[1], MAXVIEWLEN, "filename: %s", dev->file);
+   }
+   return (rc);
+}
+
+/***********************************************************************
+* tapeopen - Open tape.
+***********************************************************************/
+
+int 
+tapeopen (Device *dev)
+{
+   int rc;
+
+#ifdef DEBUGTAPE
+   fprintf (stderr, "tapeopen: ENTERED\n");
+   fprintf (stderr, "   file = %s\n", dev->file);
+#endif
+
+   if (dev->infd == DEV_NULL)
+   {
+      PUTW0 (dev, OLBIT);
+      PUTW7 (dev, IDLEBIT | ERRBIT | ABNBIT); /* Mark idle */
+      return (0);
+   }
+
+   if (!dev->realdevice)
+   {
+      /*
+      ** Check to see if a valid tape file.
+      */
+      if (tapecheck (dev) < 0)
+	 return (-1);
+   }
+   else
+      dev->sfd = fileno (dev->infd);
+
+   /*
+   ** Rewind the tape.
+   */
+
+   rc = taperewind (dev);
+   tapesetchar (dev);
+
+   PUTW0 (dev, BOTBIT | ((dev->switches & SWPROTECT) ? WPBIT : 0));
+   PUTW7 (dev, IDLEBIT); /* Mark idle */
+
+   return (rc);
+}
+
 /***********************************************************************
 * tapereset - Reset device
 ***********************************************************************/
@@ -213,17 +618,17 @@ tapereset (Device *dev)
    fprintf (stderr, "tapereset: ENTERED\n");
 #endif
 
-   if (dev->fd == DEV_NULL)
+   if (dev->infd == DEV_NULL)
    {
-      PUTW0 (dev, OLBIT | WPBIT);
+      PUTW0 (dev, OLBIT);
+      PUTW7 (dev, IDLEBIT | ERRBIT | ABNBIT ); 
    }
    else
    {
-      PUTW0 (dev, BOTBIT | (dev->info.tapeinfo.writeprotect ? WPBIT : 0));
+      PUTW0 (dev, BOTBIT | ((dev->switches & SWPROTECT) ? WPBIT : 0));
+      PUTW7 (dev, IDLEBIT); 
    }
 
-   PUTW7 (dev, IDLEBIT /*| ERRBIT | ABNBIT */); 
-
    return;
 }
 
@@ -236,10 +641,11 @@ tapedocmd (Device *dev)
 {
    Device *udev;
    int cmd;
-   int i, j;
-   int reclen, len;
-   uint32 addr, saddr;
+   int i;
+   int len;
+   uint32 addr;
    uint16 W0, W7;
+   uint16 delay;
    uint8 unit;
 
    /*
@@ -249,12 +655,12 @@ tapedocmd (Device *dev)
    W0 = GETW0 (dev);
    W7 = GETW7 (dev);
 
-   cmd = GETCMD (dev);
    unit = GETUNIT (dev);
-   saddr = addr = ((GETADDR (dev)) + 1) & 0x1FFFFE;
-   len = GETLEN(dev) /*(GETLEN (dev) + 1) & 0xFFFE*/;
+   addr = ((GETADDR (dev)) + 1) & 0x1FFFFE;
+   len = GETLEN(dev);
 
 #ifdef DEBUGTAPE
+   cmd = GETCMD (dev);
    fprintf (stderr,
 	    "tapedocmd: ENTRY: cmd = %d, unit = %02X, addr = %06X, len = %d\n",
 	    cmd, unit, addr, len);
@@ -262,7 +668,7 @@ tapedocmd (Device *dev)
 	    pcreg, wpreg, statreg);
    for (i = 0; i < 16; i+=2)
       fprintf (stderr, "   ctl[%04X] %04X\n",
-	    dev->devaddr+TPCSSTART+i, GETMEM0 (dev->devaddr+TPCSSTART+i));
+	       dev->devaddr+TPCSSTART+i, GETMEM0 (dev->devaddr+TPCSSTART+i));
 #endif
 
    /*
@@ -278,21 +684,26 @@ tapedocmd (Device *dev)
       fprintf (stderr, "tapedocmd: unit = %d: Unit not selected\n", unit);
 #endif
       /* Drive not ready */
-      W0 |= OLBIT | (dev->info.tapeinfo.writeprotect ? WPBIT : 0);
-      W7 |= ERRBIT | IDLEBIT;
+      if (W7 & INTBIT)
+	 dev->intdelay = 10;
+      W0 |= OLBIT | ((dev->switches & SWPROTECT) ? WPBIT : 0);
+      W7 |= ABNBIT | ERRBIT | IDLEBIT;
       PUTW0 (dev, W0);
       PUTW7 (dev, W7);
-      return(0);
+      return (0);
    }
    udev = &devices[i];
 
-   if (udev->fd == DEV_NULL)
+   if (udev->infd == DEV_NULL)
    {
 #ifdef DEBUGTAPE
       fprintf (stderr, "tapedocmd: unit = %d: Unit is NULL\n", unit);
 #endif
-      PUTW0 (udev, OLBIT | WPBIT);
-      PUTW7 (udev, IDLEBIT);
+      if (W7 & INTBIT)
+	 udev->intdelay = 10;
+      W7 |= ABNBIT | ERRBIT | IDLEBIT;
+      PUTW0 (udev, OLBIT );
+      PUTW7 (udev, W7);
       return (0);
    }
 
@@ -302,209 +713,92 @@ tapedocmd (Device *dev)
 
    udev->select = TRUE;
    udev->intenabled = FALSE;
-   udev->info.tapeinfo.curpos = ftell (udev->fd);
+   if (!udev->realdevice)
+      udev->info.tapeinfo.curpos = ftell (udev->infd);
+   cmd = GETCMD (udev);
 
    /*
    ** Process command
    */
 
-   switch (GETCMD (udev))
+   delay = 50;
+   switch (cmd)
    {
    case  0: /* NOP */
    case 12: /* NOP */
    case 13: /* NOP */
    case 14: /* NOP */
+      delay = 1;
       break;
 
    case 1: /* Buffer Sync */
 #ifdef DEBUGTAPE
       fprintf (stderr, "tapedocmd: BUFFSYNC: \n");
 #endif
+      delay = 10;
       break;
 
    case 2: /* Write EOF */
-#ifdef DEBUGTAPE
-      fprintf (stderr, "tapedocmd: WRITEEOF: \n");
-#endif
-      if (!udev->info.tapeinfo.writeprotect)
+      if (!(dev->switches & SWPROTECT))
       {
-
-	 tapewriteint (udev->fd, 0);
-
+	 tapeweof (udev);
 	 W0 |= EOTBIT | OLBIT;
       }
       break;
 
    case 3: /* Record skip backward */
-#ifdef DEBUGTAPE
-      fprintf (stderr, "tapedocmd: BCKSKIP: \n");
-      fprintf (stderr, "   cp0 = %ld\n", udev->info.tapeinfo.curpos);
-#endif
-      while (len)
+      if (tapebsr (udev, len) < 0)
+         return (0);
+      if (udev->info.tapeinfo.curpos == 0)
       {
-	 udev->info.tapeinfo.curpos -= 4;
-#ifdef DEBUGTAPE
-	 fprintf (stderr, "   cp1 = %ld\n", udev->info.tapeinfo.curpos);
-#endif
-
-	 if (fseek (udev->fd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
-	 {
-	    sprintf (view[0], "tapedocmd: BCKSKIP: seek failed: %s",
-		  strerror (ERRNO));
-	    sprintf (view[1], "filename: %s", udev->file);
-	    return (0);
-	 }
-	 reclen = tapereadint (udev->fd);
-#ifdef DEBUGTAPE
-	 fprintf (stderr, "   reclen = %d\n", reclen);
-#endif
-	 if (reclen > 0)
-	    udev->info.tapeinfo.curpos -= reclen + 4;
-#ifdef DEBUGTAPE
-	 fprintf (stderr, "   cp2 = %ld\n", udev->info.tapeinfo.curpos);
-#endif
-	 if (fseek (udev->fd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
-	 {
-	    sprintf (view[0], "tapedocmd: BCKSKIP: seek failed: %s",
-		  strerror (ERRNO));
-	    sprintf (view[1], "filename: %s", udev->file);
-	    return (0);
-	 }
-	 len --;
+	 W0 |= (udev->unit << 4) | BOTBIT;
+	 W7 |= COMPBIT | TPEBIT;
       }
-      PUTLEN (udev, len); 
       break;
 
    case 4: /* Binary read data */
-      reclen = tapereadint (udev->fd);
-#ifdef DEBUGTAPE
-      fprintf (stderr, "tapedocmd: READ: reclen = %d\n", reclen);
-      fprintf (stderr, "   cp = %ld\n", udev->info.tapeinfo.curpos);
-#endif
-
-      if (reclen > 0)
+      if ((i = taperead (udev, addr, len)) < 0)
       {
-	 for (j = i = 0; i < reclen;)
+	 W7 |= ERRBIT;
+	 if (i == TAPEEOF)
 	 {
-	    uint16 d, c;
-
-	    d = fgetc (udev->fd);
-	    i++;
-	    if (i < reclen)
-	    {
-	       c = fgetc (udev->fd);
-	       i++;
-	    }
-	    else
-	       c = 0xFF;
-	    d = (d << 8) | c;
-	    if (j < len)
-	    {
-	       tapeputmem (addr, d);
-	       j += 2;
-	       addr += 2;
-	    }
-	 }
-	 if (i < len)
-	 {
-	    len -= i;
-	    W0 |= EORBIT;
 	    W7 |= TPEBIT;
+	    W0 |= EOFBIT;
 	 }
-	 else len = 0;
-	 i = tapereadint (udev->fd);
-#ifdef DEBUGTAPE
-      fprintf (stderr, "   len = %d\n", len);
-      fprintf (stderr, "   traillen = %d\n", i);
-#endif
-         if (i != reclen)
+	 else if (i == TAPEEOT)
 	 {
-	    W7 |= ERRBIT;
+	    W7 |= TPEBIT;
+	    W0 |= EOTBIT;
 	 }
-
-#ifdef DEBUGTAPEDATA
-	 HEXDUMP (stderr, &memory[saddr], j, saddr);
-#endif
-	 PUTLEN (udev, len); 
       }
-      else if (reclen == 0)
+      else if (i > 0)
       {
-	 W7 |= ERRBIT | TPEBIT;
-	 W0 |= EOFBIT;
-      }
-      else if (reclen == -2)
-      {
-	 W7 |= ERRBIT | TPEBIT;
-	 W0 |= EOTBIT;
+         W0 |= EORBIT;
+	 W7 |= TPEBIT;
       }
+      delay = 100 + len;
       break;
 
    case 5: /* Record skip forward */
-#ifdef DEBUGTAPE
-      fprintf (stderr, "tapedocmd: FWDSKIP: \n");
-#endif
-
-      while (len)
+      if ((i = tapefsr (udev, len) < 0))
       {
-	 reclen = tapereadint (udev->fd);
-	 if (reclen > 0)
+	 W7 |= ERRBIT;
+	 if (i == TAPEEOF)
 	 {
-	    for (i = 0; i < reclen; i++)
-	    {
-	       fgetc (udev->fd);
-	    }
-	    i = tapereadint (udev->fd);
-	    if (i != reclen)
-	    {
-	       W7 |= ERRBIT;
-	       break;
-	    }
-	 }
-	 else if (reclen == 0)
-	 {
-	    W7 |= ERRBIT | TPEBIT;
+	    W7 |= TPEBIT;
 	    W0 |= EOFBIT;
-	    break;
 	 }
-	 else if (reclen == -2)
+	 else if (i == TAPEEOT)
 	 {
-	    W7 |= ERRBIT | TPEBIT;
+	    W7 |= TPEBIT;
 	    W0 |= EOTBIT;
-	    break;
 	 }
-	 len--;
       }
-      PUTLEN (udev, len);
       break;
 
    case 6: /* Binary write data */
-#ifdef DEBUGTAPE
-      fprintf (stderr, "tapedocmd: WRITE: \n");
-#endif
-      if (!udev->info.tapeinfo.writeprotect)
-      {
-#ifdef DEBUGTAPEDATA
-	 HEXDUMP (stderr, &memory[saddr], GETLEN(udev), saddr);
-#endif
-	 tapewriteint (udev->fd, len);
-
-	 for (i = 0; i < len; )
-	 {
-	    uint16 d;
-
-	    d = tapegetmem (addr);
-	    fputc ((d >> 8) & 0xFF, udev->fd);
-	    i++;
-	    if (i < len)
-	    {
-	       fputc (d & 0xFF, udev->fd);
-	       i++;
-	    }
-	    addr += 2;
-	 }
-	 tapewriteint (udev->fd, len);
-      }
-      PUTLEN (udev, len-i);
+      tapewrite (udev, addr, len);
+      delay = 100 + len;
       break;
 
    case 7: /* Erase */
@@ -518,22 +812,13 @@ tapedocmd (Device *dev)
 #ifdef DEBUGTAPE
       fprintf (stderr, "tapedocmd: READST: \n");
 #endif
+      if (udev->info.tapeinfo.curpos == 0)
+	 W0 |= BOTBIT;
       break;
 
    case 10: /* Rewind */
-#ifdef DEBUGTAPE
-      fprintf (stderr, "tapedocmd: REWIND: \n");
-#endif
-      udev->info.tapeinfo.curpos = 0;
-
-      if (fseek (udev->fd, 0, SEEK_SET) < 0)
-      {
-	 sprintf (view[0], "tapedocmd: REWIND: seek failed: %s",
-	       strerror (ERRNO));
-	 sprintf (view[1], "filename: %s", udev->file);
-	 return (0);
-      }
-
+      if (taperewind (udev) < 0)
+         return (0);
       W0 |= (udev->unit << 4) | BOTBIT;
       W7 |= COMPBIT | TPEBIT;
       break;
@@ -542,36 +827,31 @@ tapedocmd (Device *dev)
 #ifdef DEBUGTAPE
       fprintf (stderr, "tapedocmd: REWOFFLINE: \n");
 #endif
-      udev->info.tapeinfo.curpos = 0;
-      if (fseek (udev->fd, 0, SEEK_SET) < 0)
-      {
-	 sprintf (view[0], "tapedocmd: RESTORE: seek failed: %s",
-	       strerror (ERRNO));
-	 sprintf (view[1], "filename: %s", udev->file);
-	 return (0);
-      }
-
+      if (taperewind (udev) < 0)
+         return (0);
       W0 |= (udev->unit << 4) | BOTBIT | OLBIT;
       W7 |= COMPBIT | TPEBIT;
+      delay = 200;
       break;
-
    }
 
    if (W7 & INTBIT)
       udev->intenabled = TRUE;
    W7 |= COMPBIT | IDLEBIT;
-   W0 |= (udev->info.tapeinfo.writeprotect ? WPBIT : 0);
+   W0 |= ((dev->switches & SWPROTECT) ? WPBIT : 0);
    PUTW0 (udev, W0);
    PUTW7 (udev, W7);
 #ifdef DEBUGTAPE
-   fprintf (stderr, "returns: intenabled = %s\n",
-	 udev->intenabled ? "TRUE" : "FALSE");
+   fprintf (stderr, "returns: intenabled = %s, delay = %d\n",
+	    udev->intenabled ? "TRUE" : "FALSE", delay);
    for (i = 0; i < 16; i+=2)
       fprintf (stderr, "   ctl[%04X] %04X\n",
-	    udev->devaddr+TPCSSTART+i, GETMEM0 (udev->devaddr+TPCSSTART+i));
+	       udev->devaddr+TPCSSTART+i, GETMEM0 (udev->devaddr+TPCSSTART+i));
 #endif
 
-   return (udev->intenabled ? udev->intlvl : 0);
+   if (udev->intenabled)
+      udev->intdelay = delay;
+   return (0);
 }
 
 
@@ -582,36 +862,24 @@ tapedocmd (Device *dev)
 int
 tapeboot (Device *dev)
 {
-   run = FALSE;
-   SET_MASK (0);
-
-
-   if (model == 4 || model == 9)
+   if (bootfromrom)
    {
-      sprintf (view[0], "tapeboot: TILINE not supported on 990/%d\n", model);
-      return (-1);
+      wpreg = GETMEM0 (TPCSSTART+0xFFFC) & 0xFFFE;
+      pcreg = GETMEM0 (TPCSSTART+0xFFFE) & 0xFFFE;
+
+      PUTREG (0, 0xFFFF);
+      PUTREG (1, dev->devaddr);
+      PUTREG (2, dev->unit << 8);
    }
-
-#ifdef USETAPEROM
-
-   wpreg = GETMEM0 (TPCSSTART+0xFFFC) & 0xFFFE;
-   pcreg = GETMEM0 (TPCSSTART+0xFFFE) & 0xFFFE;
-
-#else /* !USETAPEROM */
-
+   else 
    {
       FILE *lfd = NULL;
       int lp, clp;
       int i;
+      int done;
+      int binarymode;
       int reclen;
-      char inbuf[82];
-
-      if (dev->fd == DEV_NULL)
-      {
-	 sprintf (view[0], "tapeboot: Can't boot %s device", dev->file);
-         run = FALSE;
-	 return (-1);
-      }
+      uint8 inbuf[82];
 
       clp = lp = LOADADDRESS;
 
@@ -621,32 +889,43 @@ tapeboot (Device *dev)
 	       dev->file, lp, lp);
 #endif
 
-      if (fseek (dev->fd, 0, SEEK_SET) < 0)
-      {
-	 sprintf (view[0], "tapeboot: seek failed: %s",
-	       strerror (ERRNO));
-	 sprintf (view[1], "filename: %s", dev->file);
+      if (taperewind (dev) < 0)
 	 return (-1);
-      }
-      i = 0;
 
-      while ((reclen = tapereadint (dev->fd)) == 80)
+      done = FALSE;
+      binarymode = FALSE;
+      for (i = 0; !done; i++)
       {
-	 if (fread (inbuf, 1, reclen, dev->fd) != reclen)
-	 {
-	    reclen = 0;
-	    break;
-	 }
-#ifdef DUMPLOADREC
-	 fprintf (lfd, "Record: %d\n", ++i);
-	 HEXDUMP (lfd, inbuf, 80, 0);
+#ifdef UNIX
+	 if (dev->realdevice)
+	    reclen = read (dev->sfd, inbuf, sizeof (inbuf));
+	 else
 #endif
-	 if (tapereadint (dev->fd) != reclen)
 	 {
-	    reclen = 0;
-	    break;
+	    if ((reclen = tapereadint (dev->infd)) == 80)
+	    {
+	       if (fread ((char *)inbuf, 1, reclen, dev->infd) != reclen)
+	       {
+		  reclen = 0;
+	       }
+	       else if (tapereadint (dev->infd) != reclen)
+	       {
+		  reclen = 0;
+	       }
+	    }
 	 }
-	 if (loadrec (lfd, inbuf, lp, &clp) == EOF) break;
+	 if (reclen != 80)
+	    break;
+
+	 inbuf[reclen] = 0;
+#if defined(DUMPLOADREC) && defined(DEBUGLOADER)
+	 fprintf (lfd, "Record: %d\n", i+1);
+	 HEXDUMP (lfd, inbuf, reclen, 0);
+#endif
+	 if (inbuf[0] == BINIDT_TAG)
+	    binarymode = TRUE;
+	 if (loadrec (lfd, inbuf, lp, &clp, binarymode) == EOF)
+	    done = TRUE;
       }
 
 #ifdef DEBUGLOADER
@@ -654,18 +933,15 @@ tapeboot (Device *dev)
 #endif
       if (reclen != 80)
       {
-	 sprintf (view[0], "tapeboot: read failed: %s",
-	       strerror (ERRNO));
-	 sprintf (view[1], "filename: %s", dev->file);
+	 sprintf (view[0], "Device %s: read failed: %s",
+		  dev->name, strerror (ERRNO));
+	 snprintf (view[1], MAXVIEWLEN, "filename: %s", dev->file);
          run = FALSE;
 	 return (-1);
       }
    }
 
-#endif /* USETAPEROM */
-
    run = TRUE;
 
    return (0);
 }
-
